@@ -91,7 +91,7 @@ var printTests = []printCase{
 	{"if a; then b\nelse c\nfi", "if a; then\n\tb\nelse\n\tc\nfi"},
 	samePrint("foo >&2 <f bar"),
 	samePrint("foo >&2 bar <f"),
-	{"foo >&2 bar <f bar2", "foo >&2 bar bar2 <f"},
+	{"foo >&2>/dev/null", "foo >&2 >/dev/null"},
 	{"foo <<EOF bar\nl1\nEOF", "foo bar <<EOF\nl1\nEOF"},
 	samePrint("foo <<\\\\\\\\EOF\nbar\n\\\\EOF"),
 	samePrint("foo <<\"\\EOF\"\nbar\n\\EOF"),
@@ -144,14 +144,16 @@ var printTests = []printCase{
 		"a bb\\\ncc d",
 		"a bbcc d",
 	},
-	samePrint("a \\\n\tb \\\n\tc \\\n\t;"),
-	samePrint("a=1 \\\n\tb=2 \\\n\tc=3 \\\n\t;"),
+	samePrint("a \\\n\tb \\\n\t\"c\" \\\n\t;"),
+	samePrint("a=1 \\\n\tb=2 \\\n\tc=\"3\" \\\n\t;"),
 	{
 		"a=\\\nfoo\nb=\\\n\"bar\"\nc=\\\n'baz'",
 		"a=foo\nb=\"bar\"\nc='baz'",
 	},
 
 	samePrint("if a \\\n\t; then b; fi"),
+	samePrint("a > \\\n\tfoo"),
+	samePrint("a <<< \\\n\t\"foo\""),
 	samePrint("a 'b\nb' c"),
 	samePrint("a $'b\nb' c"),
 	{
@@ -389,12 +391,21 @@ var printTests = []printCase{
 		"case $i in\n1)\n#foo \t\n;;\nesac",
 		"case $i in\n1)\n\t#foo\n\t;;\nesac",
 	},
+	{
+		"case $i in\n1)\n\t;;\n\n2)\n\t;;\nesac",
+		"case $i in\n1) ;;\n\n2) ;;\nesac",
+	},
+	{
+		"case $i\nin\n1)\n\t;;\nesac",
+		"case $i in\n1) ;;\nesac",
+	},
 	samePrint("case $i in\n1)\n\ta\n\t#b\n\t;;\nesac"),
 	samePrint("case $i in\n1) foo() { bar; } ;;\nesac"),
 	samePrint("case $i in\n1) ;; #foo\nesac"),
 	samePrint("case $i in\n#foo\nesac"),
 	samePrint("case $i in\n#before\n1) ;;\nesac"),
 	samePrint("case $i in\n#bef\n1) ;; #inl\nesac"),
+	samePrint("case $i in\n#before1\n'1') ;;\n#before2\n'2') ;;\nesac"),
 	samePrint("case $i in\n1) ;; #inl1\n2) ;; #inl2\nesac"),
 	samePrint("case $i in\n#bef\n1) #inl\n\tfoo\n\t;;\nesac"),
 	samePrint("case $i in\n1) #inl\n\t;;\nesac"),
@@ -616,6 +627,28 @@ var printTests = []printCase{
 		"`declare`",
 		"$(declare)",
 	},
+	{
+		"(\n(foo >redir))",
+		"(\n\t(foo >redir)\n)",
+	},
+	{
+		"( (foo) )",
+		"( (foo))",
+	},
+	{
+		"( (foo); bar )",
+		"(\n\t(foo)\n\tbar\n)",
+	},
+	{
+		"( ((foo++)) )",
+		"( ((foo++)))",
+	},
+	{
+		"( ((foo++)); bar )",
+		"(\n\t((foo++))\n\tbar\n)",
+	},
+	samePrint("(\n\t((foo++))\n)"),
+	samePrint("(foo && bar)"),
 }
 
 func TestPrintWeirdFormat(t *testing.T) {
